package Sim;

import java.io.File;
import java.io.IOException;

/**
 * @author Benjamin Willis
 *
 */
public class AdhocSimulation {

	/**
	 * Methods for starting a simulation.  The input requires two
	 * parameters, an input file or directory and an output directory.
	 * 
	 * @param args
	 */
	public static void main(String[] args) throws SimulationException, IOException {

		// Check for arguments
	    if ( args.length != 2 ) {
	        System.err.println( "Exactly 2 arguments are required." );
	        printUsage();
	        System.exit( 1 );
	    }
	    
	    // Open input files
	    File input = new File( args[0] );
	    File outputDir = new File( args[1] );
	
	    // Check for directioy of simulation files or 
	    // a single simulation file
	    if ( input.isDirectory() ) {
	        for ( File file : input.listFiles() ) {
	            if ( file.isFile() && file.getName().endsWith( ".txt" ) )
	                runSim( file, outputDir );
	        }
	    } else if ( input.isFile() ) {
	        runSim( input, outputDir );
	    } else {
	        System.err.println( "Input file " + input.getAbsolutePath() + " must exist." );
	        printUsage();
	        System.exit( 2 );
	    }
	}
	
    /**
     * Runs a single simulation with the given file as input and the given directory as the output
     * location.
     *
     * @param inputFile non-null File pointing to a valid file
     * @param outputDir non-null File pointing to a valid directory
     *
     * @throws IOException If there are errors reading or writing input and output files
     */
    private static void runSim( File inputFile, File outputDir )
        throws IOException {
        long startTime = System.currentTimeMillis();
        System.out.println( "Running simulation with " + inputFile.getAbsolutePath() );
        try {
            SimulationRunner runner = SimulationBuilder.buildSimulation( inputFile, outputDir );
            runner.runSimulation();
        } catch ( SimulationException e ) {
            printSimError( inputFile, e );
        } catch ( AssertionError e ) {
            printSimError( inputFile, e );
        }
        long diff = ( System.currentTimeMillis() - startTime ) / 1000;
        System.out.println( "  Finished in " + diff + " seconds" );
    }
	
    /**
     * Prints an error message saying that the Simulation's execution failed.
     */
    private static void printSimError( File inputFile, Throwable e ) {
        System.out.println( "Running simulation input " + inputFile.getAbsolutePath() +
                            " failed." );
        e.printStackTrace();
    }
    
    /**
     * Prints usage information for the class.
     */
    private static void printUsage() {
        System.err.println( "java Sim.AdhocSimulation <input> <output dir>" );
        System.err.println( " <input> may refer to a file, or a directory." );
        System.err.println( " If it refers to a directory, all .txt files are taken as inputs" );
    }
}
