package Sim;

/**
 * Packet contains data that will be sent at the network layer.
 *
 * @author Jason Winnebeck
 */
public class Packet {
	private final int source;
	private final int dest;

	private final int dataSize;

	private final PacketListener listener;

	private boolean arrived = false;
	
	private int hops = 0;
	private double distance = 0;
	/**
	 * Constructs a new Packet.
	 *
	 * @param source   source Node address
	 * @param dest     destination Node address
	 * @param dataSize size of the user data segment of this packet
	 * @param listener nullable listener to receive events.
	 */
	public Packet( int source, int dest, int dataSize, PacketListener listener ) {
		if ( dataSize < 0 )
		  throw new IllegalArgumentException( "Parameter dataSize to Packet.Packet must be positive or zero, but " + dataSize + " was given." );

		this.source = source;
		this.dest = dest;
		this.dataSize = dataSize;
		this.listener = listener;
	}

	/**
	 * Constructs a new Packet, without a listener.
	 *
	 * @param source   source Node address
	 * @param dest     destination Node address
	 * @param dataSize size of the user data segment of this packet
	 */
	public Packet( int source, int dest, int dataSize ) {
		this( source, dest, dataSize, null );
	}

	/**
	 * Tells a Packet that it has been introduced to a Node for sending.
	 *
	 * @param time Time of origination
	 */
	public void reportSent( double time ) {
		if ( listener != null )
			listener.packetOriginated( this, time );
	}

	/**
	 * Tells a Packet that it has arrived at its destination at the given time.
	 *
	 * @param time Time of arrival, must be greater than or equal to time reported in
	 *             {@link #reportSent(double)}.
	 */
	public void reportArrival( double time ) {
		if ( listener != null ){
			listener.packetArrived( this, time, arrived ); //want false first time
			listener.reportTotalHops(this,hops,distance);
		}

		arrived = true;
	}
	/**
	 * reports when this packet has gone a hop;
	 *
	 */
	public void reportHop(double d){
		distance+=d;
		hops++;
	}
	
	/**
	 * Reports true if this Packet has arrived at its destination at least once.
	 */
	public boolean hasArrived() {
		return arrived;
	}

	public int getSource() {
		return source;
	}

	public int getDest() {
		return dest;
	}

	/**
	 * Returns the size of the serialized form of this Packet in number of bytes.
	 */
	public int getSize() {
		return 8 + dataSize;
	}

	public int getDataSize() {
		return dataSize;
	}

	public String toString() {
		return "Packet from " + source + " to " + dest + " of " + dataSize + " bytes";
	}
}
