package Sim.Routing.DSR;

import Sim.HardwareFrame;
import Sim.Node;
import Sim.Packet;
import Sim.ProtocolFrame;
import Sim.Routing.AbstractRoutingAlgorithm;
import Sim.Simulation;
import Sim.SimulationEvent;
import org.gillius.Util.CheckUtils;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * DSRRoutingAlgorithm implements most of the DSR protocol.
 *
 * @author Jason Winnebeck
 */
public class DSRRoutingAlgorithm extends AbstractRoutingAlgorithm {
	private static final int    MAX_HOPS          = 10;
	private static final double RESEND_TIMEOUT    = 0.5;
	private static final int    NUM_RETRIES       =  2;
	private static final double FLOOD_EXPIRE_TIME = 1.0;
	private static final double ROUTE_REQ_TIMEOUT = 1.0; //TODO: expoential backoff as in paper

	private Set<PacketId>           floods       = new HashSet<PacketId>      ();
	private Map<Integer, Route>     routes       = new HashMap<Integer, Route>();
	private Set<Integer>            neededRoutes = new HashSet<Integer>       ();
	private List<QueueItem>         queued       = new ArrayList<QueueItem>();
	private Map<PacketId, SendData> sending      = new IdentityHashMap<PacketId, SendData>();

	private int nextId = 0;

	// ---- LOG SYSTEM
	//TODO: consider removing before release
	private static boolean doLog = false;
	private static final NumberFormat format = new DecimalFormat( "#.####" );

	/**
	 * Set to true to enable logging of a lot of debugging information to standard output.
	 */
	public static void enableLog( boolean enabled ) {
		doLog = enabled;
	}

	private void forceLog( String s ) {
		System.out.println( "[" + format.format( getTime() ) + "] Node " + myNode.getAddress() + ": " + s );
	}

	private void log( String s ) {
		if ( doLog )
			forceLog( s );
	}

	// ---- LOG SYSTEM

	/**
	 * Event when a HardwareFrame arrives destined for this Node that does not have a user Packet, or
	 * has a Packet not for this Node. The data in the frame may need to be routed.
	 *
	 * @param in non-null HardwareFrame received
	 */
	@Override
	public void arrived( HardwareFrame in ) {
		ProtocolFrame header = in.getProtocolData();
		if ( header instanceof RouteRequest ) {
			RouteRequest request = (RouteRequest) header;
			//If we haven't seen this before and it's not a loop, handle it
			if ( !floods.contains( request.getId() ) &&
			     !request.getPathTaken().containsAddress( myNode.getAddress() ) ) {
				handleRouteRequest( request );
			}

		} else if ( header instanceof DataHeader ) {
			DataHeader dataHeader = (DataHeader) header;

			log( "Got " + dataHeader );

			if ( dataHeader.isSingleHop() ) {
				handleArrivedHeader( in, dataHeader );
			} else {
				performCaching( dataHeader );
				routeData( dataHeader, in.getUserData() );
			}

		} else if ( header instanceof AckHeader ) {
			AckHeader ack = (AckHeader) header;

			//---- DEBUG
			if ( sending.containsKey( ack.getId() ) ) {
				log( "Explicit ACK for " + sending.get( ack.getId() ).frame );
			}
			//---- END DEBUG

			sending.remove( ack.getId() );

		} else {
			throw new IllegalArgumentException( "Unknown type of header: " + header );
		}
	}

	/**
	 * Handles an incoming route request, which may be intended for this node, or it may need to be
	 * forwarded.
	 */
	private void handleRouteRequest( RouteRequest request ) {
		request = request.makeRepeat( myNode.getAddress() );
		Route pathToSource = request.getPathToSource();

		if ( request.getTarget() == myNode.getAddress() ) {
			log( "Request " + request + " arrived" );
			Route pathToTarget = request.getPathTaken();
			handleNewRoute( pathToSource );
			//Don't use reliable send. Should we use rev or best route?
			//TODO: limit route spamming from incoming floods.
			routeData( new DataHeader( pathToSource, getNextId(), new RouteReply( pathToTarget ) ), null );

		} else if ( request.getHopsRemaining() > 0 ) {
			log( "Forwarding " + request );
			myNode.sendFrame( new HardwareFrame( myNode.getAddress(), HardwareFrame.BROADCAST,
			                                     request, null ) );

			handleNewRoute( pathToSource );

			floods.add( request.getId() );
			scheduleEvent( getTime() + FLOOD_EXPIRE_TIME, new FloodExpireEvent( request.getId() ) );
		}
	}

	/**
	 * Performs handling of a non-user packet protocol header arrival.
	 */
	private void handleArrivedHeader( HardwareFrame in, DataHeader dataHeader ) {
		if ( dataHeader.getNextAddress() != myNode.getAddress() )
			throw new IllegalArgumentException( "Bad route, we aren't destination" );

		//Arrived... Because this isn't terminated event, we shouldn't have a Packet
		if ( in.getUserData() != null )
			throw new IllegalArgumentException( "Routing information does not take Packet to dest" );
		myNode.sendFrame( new HardwareFrame( myNode.getAddress(), in.getSource(),
		                                     new AckHeader( dataHeader.getId() ), null ) );

		ProtocolFrame extraData = dataHeader.getExtraData();
		if ( extraData == null )
			throw new IllegalArgumentException( "Expected DSR-internal packet at this point" );

		if ( extraData instanceof RouteReply ) {
			handleRouteReply( (RouteReply) extraData );

		} else if ( extraData instanceof RouteError ) {
			handleRouteError( (RouteError) extraData );

		} else {
			throw new IllegalArgumentException( "Unknown type of extra data: " + extraData );
		}
	}

	/**
	 * Given a DataHeader that arrived, check to see if it has information in it that we can use
	 * for route caching. We can get information from route replies and route errors that are being
	 * propagated through the system.
	 *
	 * @param dataHeader non-null DataHeader
	 */
	private void performCaching( DataHeader dataHeader ) {
		ProtocolFrame extraData = dataHeader.getExtraData();
		if ( extraData != null ) {
			if ( extraData instanceof RouteReply ) {
				RouteReply reply = (RouteReply) extraData;
				handleNewRoute( reply.getRouteReply().getRouteFrom( myNode.getAddress() ) );

			} else if ( extraData instanceof RouteError ) {
				handleRouteError( (RouteError) extraData );

			} else {
				throw new IllegalArgumentException( "Unknown type of extra data: " + extraData );
			}
		}
	}

	/**
	 * Event when a HardwareFrame arrives destined for this Node that has a user Packet that is
	 * destined for this Node. The data has arrived at its final destination, so it does not need to be
	 * routed.
	 *
	 * @param in non-null HardwareFrame received
	 */
	@Override
	public void terminated( HardwareFrame in ) {
		DataHeader header = (DataHeader) in.getProtocolData();
		myNode.sendFrame( new HardwareFrame( myNode.getAddress(), in.getSource(),
		                                     new AckHeader( header.getId() ), null ) );
	}

	/**
	 * Event when a Node wants to send a new Packet. The RoutingAlgorithm should bundle the Packet in a
	 * HardwareFrame and send it through the Node.
	 *
	 * @param dest   non-null Node as the destination (packet.getDest() == dest.getAddress())
	 * @param packet user packet data to send, with a destination address not equal to the address for
	 *               the sending Node.
	 */
	@Override
	public void newPacket( Node dest, Packet packet ) {
		log( "New packet " + packet );
		sendReliable( new PacketQueueItem( packet ) );
	}

	/**
	 * Introduces a new QueueItem into the system for routing. Evokes route discovery if needed, else
	 * uses the best cached route.
	 *
	 * @param item non-null QueueItem
	 */
	private void sendReliable( QueueItem item ) {
		int target = item.getDest();

		if ( routes.containsKey( target ) ) {
			log( "Using existing route (" + routes.get( target ) + ") for " + item );
			routeData( new DataHeader( routes.get( target ), getNextId(), item.getExtraData() ),
			           item.getPacket() );

		} else {
			//Perform route discovery
			//TODO: we never give up trying to find a route...
			queued.add( item );
			if ( neededRoutes.add( target ) ) {
				//TODO: handle possible event duplication
				scheduleEvent( getTime(), new RouteRequestSendEvent( target ) );
			}
		}
	}

	/**
	 * Event when a HardwareFrame is overheard by a Node (the HardwareFrame's destination is not the
	 * Node).
	 * <p/>
	 * For DSR, this checks for passive acknowledgement
	 *
	 * @param in non-null HardwareFrame received
	 */
	@Override
	public void overheard( HardwareFrame in ) {
//		log( "Overheard " + in );

		ProtocolFrame header = in.getProtocolData();
		if ( header instanceof DataHeader ) {
			DataHeader dataHeader = (DataHeader) header;

			SendData sd = sending.get( dataHeader.getId() );
			if ( sd != null ) {
				log( "Implicit ACK for " + sd.frame );
				sending.remove( dataHeader.getId() );

			} else { //an overheard packet that isn't an implicit ACK
				ProtocolFrame extraData = dataHeader.getExtraData();
				if ( extraData != null ) {
					if ( extraData instanceof RouteError ) {
						handleRouteError( (RouteError) extraData );

					} else if ( extraData instanceof RouteReply ) {
						//TODO: handle this later
						//We choose not to because we don't have route shortening

					} else {
						throw new IllegalArgumentException( "Unknown type of extra data: " + extraData );
					}
				}
			}
		}
	}

	/**
	 * "Comparison" function to return the best of the two given routes.
	 */
	private static Route bestRoute( Route a, Route b ) {
		return ( a.numHops() < b.numHops() ) ? a : b;
	}

	/**
	 * Handles a route reply that has arrived at the node originating a route request.
	 */
	private void handleRouteReply( RouteReply reply ) {
		handleNewRoute( reply.getRouteReply() );
	}

	/**
	 * Adds a new Route and all sub-routes to the cache.
	 *
	 * @param newRoute non-null Route
	 */
	private void handleNewRoute( Route newRoute ) {
		CheckUtils.checkNull( newRoute, "newRoute" );
		if ( newRoute.getFirstHop() != myNode.getAddress() )
			throw new IllegalArgumentException( "Route not from myself?" );

		while ( newRoute.numHops() > 1 ) {
			handleNewRouteImpl( newRoute );
			newRoute = newRoute.removeLastHop();
		}
	}

	/**
	 * Adds a new Route to the cache. To be called from handleNewRoute only.
	 *
	 * @param newRoute non-null Route
	 */
	private void handleNewRouteImpl( Route newRoute ) {
		int target = newRoute.getLastHop();
		if ( target == myNode.getAddress() )
			throw new IllegalArgumentException( "Route for myself?" );

		Route oldRoute = routes.get( target );
		if ( oldRoute == null ) {
			log( "Got route " + newRoute + ", new route" );
			routes.put( target, newRoute );
			neededRoutes.remove( target );

			//Release any queued packets
			for ( Iterator<QueueItem> it = queued.iterator(); it.hasNext(); ) {
				QueueItem item = it.next();
				if ( item.getDest() == target ) {
					log( "Releasing " + item );
					routeData( new DataHeader( newRoute, getNextId(), item.getExtraData() ),
					           item.getPacket() );
					it.remove();
				}
			}
		} else {
			Route bestRoute = bestRoute( oldRoute, newRoute );
			log( "Got route " + newRoute + ", old is " + oldRoute + ", keeping " + bestRoute );
			if ( oldRoute != bestRoute )
				routes.put( target, bestRoute );
		}
	}

	/**
	 * Checks all cached routes, removing any route containing the broken link.
	 *
	 * @param error non-null RouteError
	 */
	private void handleRouteError( RouteError error ) {
		handleRouteError( error.getLinkStart(), error.getLinkEnd() );
	}

	/**
	 * Checks all cached routes, removing any route containing the broken link.
	 *
	 * @param linkStart start of the link
	 * @param linkEnd   end of the link
	 */
	private void handleRouteError( int linkStart, int linkEnd ) {
		//TODO: since links are bidirectional, we should have also removed the reciprocal!
		log( "Removing routes with broken link " + linkStart + "-" + linkEnd );
		for ( Iterator<Route> it = routes.values().iterator(); it.hasNext(); ) {
			Route route = it.next();
			if ( route.containsLink( linkStart, linkEnd ) ) {
				log( " == Removing " + route );
				it.remove();
			}
		}
	}

	/**
	 * Given a DataHeader, routes it to the next hop and schedules the appropriate resend events.
	 */
	private void routeData( DataHeader header, Packet p ) {
		if ( header.getNextAddress() != myNode.getAddress() )
			throw new IllegalArgumentException( "Bad route" );

		DataHeader next = header.makeForward();
		HardwareFrame frame = new HardwareFrame( myNode.getAddress(), next.getNextAddress(),
		                                         next, p );

		sending.put( header.getId(), new SendData( frame, next.getSecondAddress() ) );
		scheduleEvent( getTime() + RESEND_TIMEOUT, new Resender( header.getId() ) );

		myNode.sendFrame( frame );
	}

	/**
	 * Returns a unique PacketId that can be used for a new originated packet.
	 */
	private PacketId getNextId() {
		return new PacketId( myNode.getAddress(), nextId++ );
	}

	/**
	 * Encapsulates a piece of data that needs to be sent.
	 */
	private static class SendData {
		public final HardwareFrame frame;
		public final int nextHop;

		public SendData( HardwareFrame frame, int nextHop ) {
			this.frame = frame;
			this.nextHop = nextHop;
		}
	}

	/**
	 * A recurring event that keeps sending route requests for a destination until it sees that there
	 * is no longer a route needed.
	 */
	private class RouteRequestSendEvent implements SimulationEvent {
		private final int target;

		public RouteRequestSendEvent( int target ) {
			this.target = target;
		}

		public void perform( Simulation sim, double time ) {
			if ( neededRoutes.contains( target ) ) {
				RouteRequest request = new RouteRequest( getNextId(), MAX_HOPS, myNode.getAddress(), target );
				HardwareFrame frame = new HardwareFrame( myNode.getAddress(), HardwareFrame.BROADCAST,
				                                         request, null );
				log( "Originating " + request );
				myNode.sendFrame( frame );

				sim.scheduleEvent( time + ROUTE_REQ_TIMEOUT, this );
			}
		}
	}

	/**
	 * An event that when performed expires an entry in the flood cache.
	 */
	private class FloodExpireEvent implements SimulationEvent {
		private final PacketId id;

		public FloodExpireEvent( PacketId id ) {
			this.id = id;
		}

		public void perform( Simulation sim, double time ) {
			floods.remove( id );
		}
	}

	/**
	 * An event for handling resending of a SendData for a specific PacketID.
	 */
	private class Resender implements SimulationEvent {
		private final PacketId id;
		private int retriesLeft = NUM_RETRIES;

		public Resender( PacketId id ) {
			this.id = id;
		}

		public void perform( Simulation sim, double time ) {
			SendData sd = sending.get( id );
			if ( sd != null ) {
				if ( retriesLeft > 0 ) {
					//Then we actually do want to do the resend
					log( "Resending " + sd.frame );
					++resends; //testing support only!
					myNode.sendFrame( sd.frame );
					sim.scheduleEvent( sim.getTime() + RESEND_TIMEOUT, this );
					--retriesLeft;

				} else {
					log( "Out of retries, dropping " + sd.frame );

					//Drop all frames waiting to be sent to bad destination
					for ( Iterator<SendData> it = sending.values().iterator(); it.hasNext(); ) {
						SendData sendData = it.next();
						if ( sendData.frame.getDest() == sd.frame.getDest() )
							it.remove();
					}

					//Send back a route error if needed, and update our caches
					if ( sd.frame.getUserData() != null ) {
						int source = sd.frame.getUserData().getSource();
						handleRouteError( myNode.getAddress(), sd.frame.getDest() );
						if ( source != myNode.getAddress() ) {
							RouteError error = new RouteError( myNode.getAddress(), sd.frame.getDest() );
							sendReliable( new ProtocolFrameQueueItem( source, error ) );
						} else {
							//This is an error from our side
						}
					}
				}
			}
		}
	}

	//Statistics, testing support
	int resends = 0;

	int getSendingSize() {
		return sending.size();
	}

	int getQueued() {
		return queued.size();
	}

	int getNeededRoues() {
		return neededRoutes.size();
	}
}
