package Sim.Routing.DSR;

import Sim.ProtocolFrame;

/**
 * DataHeader represents a packet with an explicit route. A DataHeader is put onto any routed
 * (opposed to flooded) packet, and can contain a user data Packet or a nexted ProtocolFrame as
 * extra data.
 *
 * @author Jason Winnebeck
 */
public class DataHeader implements ProtocolFrame {
	private final Route         route;
	private final PacketId      id;
	private final ProtocolFrame extraData;

	/**
	 * Constructs a new DataHeader that is for routing a user packet.
	 *
	 * @param route non-null Route to take
	 * @param id    non-null id to be used throughout this packet's lifetime
	 */
	public DataHeader( Route route, PacketId id ) {
		this.route = route;
		this.id    = id;
		extraData  = null;
	}

	/**
	 * Constructs a new DataHeader to contain a nested ProtocolFrame.
	 *
	 * @param route     non-null Route to take
	 * @param id        non-null id to be used throughout this packet's lifetime
	 * @param extraData nullable ProtocolFrame (if null, works like
	 *                  {@link #DataHeader( Route, PacketId ) }
	 */
	public DataHeader( Route route, PacketId id, ProtocolFrame extraData ) {
		this.route     = route;
		this.id        = id;
		this.extraData = extraData;
	}

	/**
	 * Returns the remaining portion of the route to the destination.
	 *
	 * @return non-null Route
	 */
	public Route getRouteRemaining() {
		return route;
	}

	public PacketId getId() {
		return id;
	}

	/**
	 * DataHeaders can carry additional data for use by DSR itself, and if a DSR-internal packet is
	 * being routed, it is contained as extra data.
	 *
	 * @return nullable ProtocolFrame
	 */
	public ProtocolFrame getExtraData() {
		return extraData;
	}

	/**
	 * Returns true if the contained route has only a single hop (usually this means the packet is at
	 * the destination).
	 */
	public boolean isSingleHop() {
		return route.numHops() == 1;
	}

	/**
	 * Returns the next hop.
	 */
	public int getNextAddress() {
		return route.getFirstHop();
	}

	/**
	 * Returns the address of the hop after this one, or an invalid address less than 0 if there is
	 * no next hop.
	 */
	public int getSecondAddress() {
		return isSingleHop() ? -1 : route.getSecondHop();
	}

	/**
	 * Constructs a new DataHeader that is a forwarded version of this header.
	 */
	public DataHeader makeForward() {
		return new DataHeader( route.removeFirstHop(), id, extraData );
	}

	/**
	 * Returns the "serialized" size of this portion of the packet.
	 */
	public int getSize() {
		return route.getSize() + id.getSize() + ( extraData == null ? 0 : extraData.getSize() );
	}

	public String toString() {
		return "DataHeader " + route;
	}
}
