package Sim.Routing.DSR;

import org.gillius.Util.CheckUtils;

/**
 * RouteRequest is a flooded protocol header to request a route from the originator to a target.
 *
 * @author Jason Winnebeck
 */
public class RouteRequest extends FloodingHeader {
	private final Route    pathTaken;
	private final int      target;
	private final PacketId id;

	private final RouteError error;

	public RouteRequest( PacketId id, int hopsRemaining, int originator, int target,
	                     RouteError error ) {
		super( hopsRemaining );
		CheckUtils.checkNull( id, "id" );
		if ( hopsRemaining <= 0 )
		  throw new IllegalArgumentException( "Parameter hopsRemaining to RouteRequest.RouteRequest must be positive, but " + hopsRemaining + " was given." );
		if ( originator == target )
			throw new IllegalArgumentException( "Cannot have route request for self" );

		pathTaken   = new Route( originator );
		this.target = target;
		this.id     = id;
		this.error  = error;
	}

	public RouteRequest( PacketId id, int hopsRemaining, int originator, int target ) {
		this( id, hopsRemaining, originator, target, null );
	}

	public RouteRequest( PacketId id, int hopsRemaining, Route pathTaken, int target,
	                     RouteError error ) {
		super( hopsRemaining );

		this.pathTaken = pathTaken;
		this.target = target;
		this.id = id;
		this.error = error;
	}

	public int getOriginator() {
		return id.getOriginator();
	}

	public PacketId getId() {
		return id;
	}

	public int getTarget() {
		return target;
	}

	public boolean shouldRepeat( int myAddress ) {
		return getHopsRemaining() > 0 && !pathTaken.containsAddress( myAddress );
	}

	/**
	 * Constructs a new RouteRequest with the address given appended to the path, suitable for
	 * forwarding.
	 */
	public RouteRequest makeRepeat( int myAddress ) {
		return new RouteRequest( id, getHopsRemaining() - 1, pathTaken.addHop( myAddress ), target,
		                         error );
	}

	public Route getPathToSource() {
		return pathTaken.reverse();
	}

	/**
	 * Terminates the RouteRequest, returning the complete route obtained.
	 * @return non-null Route
	 */
	public Route terminate() {
		return pathTaken.addHop( target );
	}

	public Route getPathTaken() {
		return pathTaken;
	}

	/**
	 * Returns the RouteError associated with this RouteRequest, if there is one, or null if there is
	 * not. Nodes should not respond to a RouteRequest with a route from their cache that contains a
	 * link found in this RouteError.
	 * <p/>
	 * TODO: currently this is not used.
	 */
	public RouteError getError() {
		return error;
	}

	/**
	 * Returns the "serialized" size of this portion of the packet.
	 */
	@Override
	public int getSize() {
		return super.getSize() +
		       ( pathTaken.getSize() - 4 ) + //don't need originator; it's in ID
		       id.getSize() +
				   1 + ( ( error != null ) ? error.getSize() : 0 ); //1 byte existance header
	}

	public String toString() {
		return "RouteRequest (ID " + id + ") to " + target + ", Path: " + pathTaken;
	}
}
