package Sim.Routing;

import Sim.HardwareFrame;
import Sim.Node;
import Sim.Packet;
import Sim.Simulation;

/**
 * RoutingAlgorithm specifies the interface for a routing algorithm for a Node. Frames and new
 * Packets are passed from the Node to the RoutingAlgorithm, and the RoutingAlgorithm is responsible
 * for queuing a HardwareFrame at the Node in order to forward the Packet towards its destination.
 * <p/>
 * In order to facilitate easy dynamic loading, all implementations of RoutingAlgorithm must have
 * a default constructor.
 *
 * @author Jason Winnebeck
 */
public interface RoutingAlgorithm {
	/**
	 * Sets the simulation for this RoutingAlgorithm. This method must be called immediately after
	 * construction and before any other methods.
	 *
	 * @param sim non-null Simulation
	 */
	public void setSimulation( Simulation sim );

	/**
	 * Event called when a Node is given a RemoteAlgorithm to route packets for it.
	 *
	 * @param us non-null Node
	 */
	public void attached( Node us );

	/**
	 * Event when a HardwareFrame arrives destined for this Node that does not have a user Packet, or
	 * has a Packet not for this Node. The data in the frame may need to be routed.
	 *
	 * @param in non-null HardwareFrame received
	 */
	public void arrived( HardwareFrame in );

	/**
	 * Event when a HardwareFrame arrives destined for this Node that has a user Packet that is
	 * destined for this Node. The data has arrived at its final destination, so it does not need to
	 * be routed.
	 *
	 * @param in non-null HardwareFrame received
	 */
	public void terminated( HardwareFrame in );

	/**
	 * Event when a HardwareFrame is overheard by a Node (the HardwareFrame's destination is not the
	 * Node).
	 *
	 * @param in non-null HardwareFrame received
	 */
	public void overheard( HardwareFrame in );

	/**
	 * Event when a Node hears its own HardwareFrame (which is guaranteed to happen after a send
	 * completes).
	 *
	 * @param in non-null HardwareFrame received whose source equals the "us" address
	 */
	public void loopback( HardwareFrame in );

	/**
	 * Event when a Node wants to send a new Packet. The RoutingAlgorithm should bundle the Packet in
	 * a HardwareFrame and send it through the Node.
	 *
	 * @param dest   non-null Node as the destination (packet.getDest() == dest.getAddress())
	 * @param packet user packet data to send, with a destination address not equal to the address
	 *               for the sending Node.
	 */
	public void newPacket( Node dest, Packet packet );
}
