package Sim;

import org.gillius.Util.CheckUtils;

import java.util.Iterator;
import java.util.PriorityQueue;
import java.util.Random;

/**
 * Simulation is the core of the discrete event simulation. It holds the current simulation time and
 * all of the events. Events are scheduled through the {@link #scheduleEvent(double,
 * SimulationEvent)} method.
 * <p/>
 * The Simulation also provides the common source of randomness, so that runs of the same Simulation
 * can be reproducable.
 *
 * @author Jason Winnebeck
 */
public class Simulation {
	private double time = 0.0;
	private String name = "Sim";

	private final PriorityQueue<EventEntry> events = new PriorityQueue<EventEntry>();
	private final Random rng;

	private final SimulationParameters params;

	private final World world;

	/**
	 * Constructs a new Simulation with the given parameters.
	 *
	 * @param p non-null SimulationParameters
	 */
	public Simulation( SimulationParameters p ) {
		CheckUtils.checkNull( p, "p" );

		this.params = p;
		rng = new Random( p.getRandomSeed() );

		world = new World( p.getNoiseRate(), rng );
	}

	/**
	 * Returns the World associated with this Simulation.
	 *
	 * @return non-null World
	 */
	public World getWorld() {
		return world;
	}

	/**
	 * Returns the source of randomness for this Simulation.
	 */
	public Random getRng() {
		return rng;
	}

	/**
	 * Returns global simulation parameters that are set at the beginning of the simulation and
	 * never change throughout the execution.
	 */
	public SimulationParameters getParams() {
		return params;
	}

	/**
	 * Returns the name of this Simulation. The default name for a Simulation is "Sim".
	 */
	public String getName() {
		return name;
	}

	/**
	 * Sets this Simulation's name
	 */
	public void setName( String name ) {
		this.name = name;
	}

	/**
	 * Performs the next event in this Simulation. After processing a single event, returns true or
	 * false depending if there are more events. If there were no events to process, then false is
	 * returned and the simulation time is not modified.
	 *
	 * @return true if there are any more events to process, or false otherwise.
	 */
	public boolean next() {
		EventEntry nextEvent = events.poll();

		if ( nextEvent != null ) {
			assert ( time <= nextEvent.eventTime );
			time = nextEvent.eventTime;
			nextEvent.event.perform( this, time );
		}

		return !events.isEmpty();
	}

	/**
	 * Returns the current simulation time, which can stand still but never go backwards. A
	 * Simulation starts with the time equal to 0.
	 */
	public double getTime() {
		return time;
	}

	/**
	 * Cancels all events in the Simulation that match a certain type.
	 *
	 * @param eventType non-null event type class
	 */
	public void cancelEvents( Class eventType ) {
		CheckUtils.checkNull( eventType, "eventType" );

		for ( Iterator iterator = events.iterator(); iterator.hasNext(); ) {
			EventEntry entry = (EventEntry) iterator.next();
			if ( entry.event.getClass().equals( eventType ) )
				iterator.remove();
		}
	}

	/**
	 * Schedules a new event to take place in the future. Once an event has been scheduled, it
	 * cannot be cancelled or rescheduled.
	 *
	 * @param t     event time greater than or equal to the current simulation time
	 * @param event non-null {@link SimulationEvent}
	 */
	public void scheduleEvent( double t, SimulationEvent event ) {
		CheckUtils.checkNull( event, "event" );
		if ( t < time )
			throw new IllegalArgumentException( "Cannot schedule an event in the past" );

		events.offer( new EventEntry( t, event ) );
	}

	/**
	 * Wraps around en event to attach an execution timestamp with it, for use in the priority
	 * queue. An EventEntry is comparable based solely on its eventTime.
	 */
	private static class EventEntry implements Comparable<EventEntry> {
		final double eventTime;
		final SimulationEvent event;

		public EventEntry( double eventTime, SimulationEvent event ) {
			this.eventTime = eventTime;
			this.event = event;
		}

		/**
		 * Compares EventEntries by their event time.
		 *
		 * @param eventEntry the object to be compared.
		 *
		 * @return a negative integer, zero, or a positive integer as this object is less than,
		 *         equal to, or greater than the specified object.
		 */
		public int compareTo( EventEntry eventEntry ) {
			return Double.compare( eventTime, eventEntry.eventTime );
		}
	}
}
