package Sim;

/**
 * SimulationParameters holds global {@link Simulation} parameters.
 *
 * @author Jason Winnebeck
 */
public class SimulationParameters {
	private final double idlePowerUse;
	private final double transmitPowerUse;
	private final double transferRate;
	private final double noiseRate;

	private final double simDuration;

	private final int randomSeed;

	/**
	 * Constructs a new SimulationParameters
	 *
	 * @param idlePowerUse     amount of power used per second while idling
	 * @param transmitPowerUse amount of power used per second while transmitting
	 * @param transferRate     wireless network transfer speed in bits per second
	 * @param noiseRate        percentage of packets lost due to noise at the receiver
	 * @param simDuration      duration of simulation in seconds
	 * @param randomSeed       random number generator seed
	 */
	public SimulationParameters( double idlePowerUse, double transmitPowerUse, double transferRate,
	                             double noiseRate, double simDuration, int randomSeed ) {
		if ( transferRate < 0 )
		  throw new IllegalArgumentException( "Parameter transferRate to SimulationParameters.SimulationParameters must be positive or zero, but " + transferRate + " was given." );
		if ( transmitPowerUse < 0 )
		  throw new IllegalArgumentException( "Parameter transmitPowerUse to SimulationParameters.SimulationParameters must be positive or zero, but " + transmitPowerUse + " was given." );
		if ( idlePowerUse < 0 )
		  throw new IllegalArgumentException( "Parameter idlePowerUse to SimulationParameters.SimulationParameters must be positive or zero, but " + idlePowerUse + " was given." );
		if ( noiseRate < 0.0 || noiseRate > 1.0 )
		  throw new IllegalArgumentException( "Parameter noiseRate to SimulationParameters.SimulationParameters must be within the range 0.0-1.0, inclusive; its value is " + noiseRate + '.' );
		if ( simDuration < 0 )
			throw new IllegalArgumentException( "Parameter simDuration to SimulationParameters.SimulationParameters must be positive or zero, but " + simDuration + " was given." );

		this.idlePowerUse = idlePowerUse;
		this.transmitPowerUse = transmitPowerUse;
		this.transferRate = transferRate;
		this.noiseRate = noiseRate;
		this.simDuration = simDuration;
		this.randomSeed = randomSeed;
	}

	/**
	 * Amount of power used per second while idling.
	 */
	public double getIdlePowerUse() {
		return idlePowerUse;
	}

	/**
	 * Amount of power used per second while transmitting.
	 */
	public double getTransmitPowerUse() {
		return transmitPowerUse;
	}

	/**
	 * Wireless network transfer speed in bits per second.
	 */
	public double getTransferRate() {
		return transferRate;
	}

	/**
	 * Percentage of packets lost due to noise at the receiver.
	 */
	public double getNoiseRate() {
		return noiseRate;
	}

	/**
	 * Gets the duration of simulation in seconds.
	 */
	public double getSimDuration() {
		return simDuration;
	}

	/**
	 * Gets the random number generator seed.
	 */
	public int getRandomSeed() {
		return randomSeed;
	}
}
