package Sim;

import edu.rit.numeric.prob.BernoulliPrng;
import edu.rit.numeric.prob.BooleanPrng;
import org.gillius.Util.CheckUtils;

import java.util.HashMap;
import java.util.Map;
import java.util.Random;
import java.util.List;
import java.util.ArrayList;

/**
 * World implements a world of Nodes. The World also simulates the physical transmission of packets,
 * and whether a packet is lost due to noise.
 *
 * @author Jason Winnebeck
 */
public class World {
	private final Map<Integer, Node> nodes = new HashMap<Integer, Node>();
	private final List<Node> nodesList = new ArrayList<Node>();

	private final BooleanPrng noisePrng;

	/**
	 * Constructs a new World with the given amount of noise. noiseRatio percent of packets will be
	 * lost on average. The loss occurs at the receiving end, if it does occur, so if a Node sends a
	 * packet, it is possible for some of its neighbors to receive it and some to not receive it.
	 *
	 * @param noiseRatio a percentage from 0 (no loss) to 1 (total loss)
	 * @param rng        non-null source of randomness to use for packet loss
	 */
	public World( double noiseRatio, Random rng ) {
		this( new BernoulliPrng( rng, noiseRatio ) );
	}

	/**
	 * Constructs a new World, using the given BooleanPrng to determine packet loss. The loss occurs at
	 * the receiving end, if it does occur, so if a Node sends a packet, it is possible for some of its
	 * neighbors to receive it and some to not receive it.
	 *
	 * @param noisePrng a percentage from 0 (no loss) to 1 (total loss)
	 */
	public World( BooleanPrng noisePrng ) {
		CheckUtils.checkNull( noisePrng, "noisePrng" );
		this.noisePrng = noisePrng;
	}

	/**
	 * Adds a Node into the World that is capable of receiving frames. There cannot be more than one
	 * Node with the same address in a World.
	 *
	 * @param n non-null Node with unique (to this World) address
	 */
	public void addNode( Node n ) {
		CheckUtils.checkNull( n, "n" );

		if ( nodes.containsKey( n.getAddress() ) ) {
			throw new IllegalArgumentException( "Node with address " + n.getAddress() +
			                                    " already exists in World" );
		}

		nodes.put( n.getAddress(), n );
		nodesList.add( n );
	}

	/**
	 * Returns the Node whose address matches the address given, if it is part of this World.
	 *
	 * @param address address of Node to return
	 *
	 * @return non-null if a Node with the address exists
	 */
	public Node getNode( int address ) {
		return nodes.get( address );
	}

	/**
	 * Returns a Node in the World at random, with an equal probability of returning any specific
	 * Node.
	 *
	 * @param rng non-null source of randomness, from which a single integer will be pulled.
	 *
	 * @return non-null Node in this World
	 */
	public Node getRandomNode( Random rng ) {
		return nodesList.get( rng.nextInt( nodesList.size() ) );
	}

	/**
	 * Removes a Node (by giving its unique address) from the World. That Node will no longer receive
	 * any frames.
	 *
	 * @param address address of Node to remove
	 *
	 * @return true if a Node was removed (address was valid)
	 */
	public boolean removeNode( int address ) {
		Node removed = nodes.remove( address );
		if ( removed != null )
			nodesList.remove( removed );
		return removed != null;
	}

	/**
	 * Sends a frame to all Nodes within listening range of the transmission point given.
	 *
	 * @param frame             non-null HardwareFrame
	 * @param transmissionPoint non-null Location
	 * @param distance          distance in which frame can be received.
	 */
	public void sendFrame( HardwareFrame frame, Location transmissionPoint, double distance ) {
		CheckUtils.checkNull( frame, "frame" );
		CheckUtils.checkNull( transmissionPoint, "transmissionPoint" );

		final double d2 = distance * distance;

		//Inefficently scan all Nodes. It's not worth building a BSP or quadtree for this
		for ( Node node : nodes.values() ) {
			if ( frame.getSource() == node.getAddress() ) {
				node.receiveFrame( frame );

			} else if ( transmissionPoint.getQuadrance( node.getLocation() ) <= d2 ) {
				if ( !noisePrng.next() ) { //if not dropped due to noise
					node.receiveFrame( frame );
				}
			}
		}
	}
}
