package Sim.Routing.DSR;

import Sim.FixedLocationManager;
import Sim.Location;
import Sim.LocationManager;
import Sim.Node;
import Sim.Simulation;
import Sim.SimulationParameters;
import Sim.Packet;
import Sim.SimpleLocationManager;
import Sim.NodeMonitor;
import Sim.SimulationEvent;
import junit.framework.TestCase;

import java.util.List;
import java.util.ArrayList;

/**
 * TestDSR
 *
 * @author Jason Winnebeck
 */
public class TestDSR extends TestCase {
	private Simulation sim;
	private List<DSRRoutingAlgorithm> algs = new ArrayList<DSRRoutingAlgorithm>();

	private SimpleLocationManager moverLoc;
	private Node mover;

	/**
	 * This method is called before a test is executed.
	 */
	@Override
	protected void setUp() throws Exception {
		super.setUp();

		SimulationParameters params = new SimulationParameters( 0, 0, 100000.0, 0.0, 100.0, 200 );

		sim = new Simulation( params );

		makeNode( 0,  0, 0 );
		makeNode( 1, 10, 0 );
		makeNode( 2, 20, 0 );
		makeNode( 3, 30, 0 );
		makeNode( 4, 40, 0 );

		DSRRoutingAlgorithm.enableLog( true );

		moverLoc = new SimpleLocationManager( new Location( 35, 0 ) );
		DSRRoutingAlgorithm alg = new DSRRoutingAlgorithm();
		alg.setSimulation( sim );

		mover = new Node( 10, sim, moverLoc, alg, 11.0, new NodeMonitor() );
	}

	private Node makeNode( int addr, double x, double y ) {
		LocationManager lm = new FixedLocationManager( new Location( x, y ) );
		DSRRoutingAlgorithm alg = new DSRRoutingAlgorithm();
		algs.add( alg );
		alg.setSimulation( sim );
		Node ret = new Node( addr, sim, lm, alg, 11.0, new NodeMonitor() );
		sim.getWorld().addNode( ret );
		return ret;
	}

	public void testDSR() {
		checkPacketArrive( new Packet( 0, 4, 100 ), true, true );

		checkStats();

		checkPacketArrive( new Packet( 0, 4, 50 ), true, true );

		checkStats();
	}

	private void checkStats() {
		for ( DSRRoutingAlgorithm alg : algs ) {
			assertEquals( 0, alg.getSendingSize() );
			assertEquals( 0, alg.resends );
			assertEquals( 0, alg.getQueued() );
			assertEquals( 0, alg.getNeededRoues() );
		}
	}

	public void testRouteDiscoveryRetransmission() {
		moverLoc.setLocation( new Location( 450, 0 ) );
		sim.getWorld().addNode( mover );

		sim.scheduleEvent( 1.1, new SimulationEvent() {
			public void perform( Simulation sim, double time ) {
				System.out.println( "Moving Node in range" );
				moverLoc.setLocation( new Location( 45, 0 ) );
			}
		});

		checkPacketArrive( new Packet( 0, 10, 200 ), true, true );
		checkStats();
	}

	public void testRouteMaintenance() {
		sim.getWorld().addNode( mover );

		checkPacketArrive( new Packet( 1, 10, 20 ), true, false );

		moverLoc.setLocation( new Location( -5, 0 ) ); //Change location

		//The first send after the move should fail because the route is bad.
		checkPacketArrive( new Packet( 1, 10, 20 ), false, true );

		//The second send of the same packet should work because the bad route should have been cleared.
		checkPacketArrive( new Packet( 1, 10, 20 ), true, true );
	}

	public void testRouteErrorFromOriginator() {
		moverLoc.setLocation( new Location( 45, 0 ) );
		sim.getWorld().addNode( mover );

		checkPacketArrive( new Packet( 4, 10, 20 ), true, false );

		moverLoc.setLocation( new Location( -5, 0 ) ); //Change location

		//The first send after the move should fail because the route is bad.
		checkPacketArrive( new Packet( 4, 10, 20 ), false, true );
	}

	private void checkPacketArrive( Packet packet, boolean shouldArrive, boolean log ) {
		DSRRoutingAlgorithm.enableLog( log );
		send( packet );
		while ( sim.next() ) {}
		assertEquals( shouldArrive, packet.hasArrived() );
	}

	private void send( Packet packet ) {
		sim.getWorld().getNode( packet.getSource() ).sendNewPacket( packet );
	}
}
